/**********************************************************************
 **                                                                  **
 **                           waypoint.c                             **
 **                                                                  **
 **  Written by:  Yoichiro Endo and Jonathan F. Diaz                 **
 **                                                                  **
 **  Copyright 2000 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: waypoint.c,v 1.1.1.1 2008/07/14 16:44:22 endo Exp $ */

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include <memory.h>
#include <malloc.h>
#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>

#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>

#include <Xm/ArrowBG.h>
#include <Xm/BulletinB.h>
#include <Xm/CascadeB.h>   
#include <Xm/DialogS.h>
#include <Xm/DrawingA.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/List.h>
#include <Xm/MainW.h>
#include <Xm/MessageB.h>
#include <Xm/PanedW.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/ScrollBar.h>
#include <Xm/ScrolledW.h>
#include <Xm/Separator.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>

#include "gt_sim.h"
#include "gt_scale.h"
#include "gt_console_db.h"
#include "draw.h"
#include "file_utils.h"
#include "waypoint.h"
#include "version.h"
#include "EventLogging.h"

static Widget ask_save_ovl_dialog;
static Widget save_ovl_as_dialog;

#define XSTRING(str) XmStringCreateLtoR(str, XmSTRING_DEFAULT_CHARSET)

void save_ovlname_to_waypoint_file(char *filename);

/****************************************************************************
* create_new_ovl_for_waypoints                                              *
*                                                                           *
* This function create an overlay file which contains WAYPOINT-FILE tag.    *
****************************************************************************/
void create_new_ovl_for_waypoints(char *filename)
{
  // check to see if the filename needs .ovl extention.
  int len = strlen(filename);
  if (len <= 0) quit_mlab();
  if (len < 5 || strcmp(&filename[len - 4], ".ovl") != 0)
  {
    // make a copy of the filename with the extension
    char *tmp = (char *) malloc(len + 5);

    strcpy(tmp, filename);
    strcpy(&tmp[len], ".ovl");
    free(filename);
    filename = tmp;
  }

  // Get current time.  
  char month[100];
  char date[100];
  char year[100];
  const time_t cur_time = time(NULL);
  tm local_time = *localtime(&cur_time);
  strftime(month,sizeof(month), "%m", &local_time);
  strftime(date,sizeof(date), "%d", &local_time);
  strftime(year,sizeof(year), "%Y", &local_time);

  // Open the waypoint file generated by the waypoint-creation function.
  FILE *wp_file;
  if (!(wp_file = fopen(waypoints_filename, "r")))
  {
    warn_userf("Cannot read waypoints file \"%s\".", waypoints_filename);
    quit_mlab();
  }

  // Generate *.wpt file which contains the information of
  // the waypoint.
  FILE *wpt_file;
  char *wpt_filename = strdup(filename);
  strcpy(wpt_filename+strlen(wpt_filename)-3, "wpt");

  if (!(wpt_file = fopen(wpt_filename, "w")))
  {
    fprintf(stderr, "Internal Error: Cannot open \"%s\".\n", wpt_filename);
    fclose(wp_file);
    quit_mlab();
  }

  // Write the header in the *.wpt file.
  fprintf(wpt_file, "-- This waypoint file was generated by mlab v%s on %s/%s/%s.\n\n",
	  version_str, month, date, year);
  fprintf(wpt_file, "MISSION-AREA %.2f %.2f\n\n",
	  mission_area_width_meters, mission_area_height_meters);
  fprintf(wpt_file, "CONTROL MEASURES:\n");

  // Now, add the copy the waypoints in the *.wpt file.
  int count = 1;
  float ppx, ppy;
  double pp_size = 10*meters_per_pixel;
  char buf[1024];

  while (fscanf(wp_file, "%s", buf) != EOF)
  {
      if (!strcmp(buf, "waypoint"))
      {
	  fscanf(wp_file,"%f", &ppx);
	  fscanf(wp_file,"%f", &ppy);

	  fprintf(wpt_file, "PP \"Waypoint #%d\" %.2f %.2f %.2f\n",
		  count, ppx, ppy, pp_size);
	  count++;
      }
  }

  // Close the files.
  fclose(wpt_file);
  fclose(wp_file);

  // Open a temp file to copy the overlay.
  char tmpfilename[4096];
  sprintf(tmpfilename, "/tmp/tmp-%s%s%s-%d.ovl", month, date, year, getpid());

  FILE *ovl_file, *tmp_file;
  char *tmp_filename = strdup(tmpfilename);
  char line[4096];

  if (!(ovl_file = fopen(ovl_filename, "r")))
  {
    fprintf(stderr, "Internal Error: Cannot read the overlay \"%s\".\n", 
	    ovl_filename);
    quit_mlab();
  }

  if (!(tmp_file = fopen(tmp_filename, "a")))
  {
    warn_userf("Internal Error: Cannot write to the temporary file \"%s\".", 
	    tmp_filename);
    fclose(ovl_file);
    quit_mlab();
  }

  // Copy the overlay untill it reaches EOF or the old waypoint file tag.
  while (feof(ovl_file) == 0)
  {
    if (fgets(line, 4096, ovl_file) != NULL)
    {
      if (strncmp(line, "WAYPOINT-FILE", 13) == 0)
      {
	// Old waypoint file tag found. Quit copying. Assuming that
	// WAYPOINT-FILE is the last CONTROL-MEASURE in the overlay.
	break;
      }
      else
      {
	// Copy it to the temp file.
	fprintf(tmp_file, "%s", line);
      }
    }
  }
  fclose(ovl_file);

  // Append the waypoint file tag to the temp file.
  char wpt_filename_str[4096];
  remove_directory(wpt_filename, wpt_filename_str);
  fprintf(tmp_file, "WAYPOINT-FILE \"%s\" -- Do not write below this line\n", wpt_filename_str);
  fprintf(tmp_file, "-- Waypoint File Tag Added by mlab v%s on %s/%s/%s\n",
	  version_str, month, date, year);
  fclose(tmp_file);

  // Now, copy this temp file to <filename> which the user specified.
  char cmd[4096];
  sprintf(cmd, "cp %s %s", tmpfilename, filename);
  system(cmd);

  gEventLogging->log("%s created", filename);

  // Delete the temp file.
  unlink(tmp_filename);

  // Save the overlay file name.
  save_ovlname_to_waypoint_file(filename);

  // Quit mlab.
  quit_mlab();
}

/****************************************************************************
* overwrite_yes_cb                                                          *
*                                                                           *
* This is a callback function that invokes create_new_ovl_for_wapoint().    *
****************************************************************************/
void overwrite_yes_cb(Widget dialog, XtPointer client_data, XtPointer call_data)
{
    char *filename = (char *)client_data;

    gEventLogging->log("File exists. Overwrite it? = Yes");
    gEventLogging->end("File exists. Overwrite it?");

    if (filename)
    {
	create_new_ovl_for_waypoints(filename);
    }
    else
    {
	warn_userf("Internal Error: Overlay not saved.");
	quit_mlab();
    }
}

/****************************************************************************
* overwrite_no_cb                                                           *
*                                                                           *
* This is a callback function that gets ivoked when the user chooses "No"   *
* in ask_overwrite_ovl_dialog.                                              *
****************************************************************************/
void overwrite_no_cb(Widget dialog, XtPointer client_data, XtPointer call_data)
{
    char *filename = NULL;

    gEventLogging->log("File exists. Overwrite it? = NO");
    gEventLogging->end("File exists. Overwrite it?");

    filename = (char *)client_data;

    Widget text = XmFileSelectionBoxGetChild(save_ovl_as_dialog, XmDIALOG_TEXT);
    XtVaSetValues (text, XmNvalue, filename, NULL);

    Widget list = XmFileSelectionBoxGetChild(save_ovl_as_dialog, XmDIALOG_LIST);
    XmString filename_ = XSTRING(filename);
    XmListSelectItem(list, filename_, FALSE);
    XmStringFree(filename_);

    XtManageChild(save_ovl_as_dialog);
    gEventLogging->start("Save Overlay As");
}

/****************************************************************************
* save_ovl_as_cb                                                            *
*                                                                           *
* This is a callback function that gets ivoked when the user chooses "save  *
* as" from ask_overwrite_ovl_dialog.                                        *
****************************************************************************/
void save_ovl_as_cb(Widget dialog, XtPointer client_data, XtPointer call_data)
{
    char *filename = NULL;

    gEventLogging->log("Do you want to save the waypoints in the overlay file? = Save As");
    gEventLogging->end("Do you want to save the waypoints in the overlay file?");

    filename = (char *)client_data;

    Widget text = XmFileSelectionBoxGetChild(save_ovl_as_dialog, XmDIALOG_TEXT);
    XtVaSetValues (text, XmNvalue, filename, NULL);

    Widget list = XmFileSelectionBoxGetChild(save_ovl_as_dialog, XmDIALOG_LIST);
    XmString filename_ = XSTRING(filename);
    XmListSelectItem(list, filename_, FALSE);
    XmStringFree(filename_);

    XtManageChild(save_ovl_as_dialog);
    gEventLogging->start("Save Overlay As");
}

/****************************************************************************
* save_ovl_cancel_cb                                                        *
*                                                                           *
* This is a callback function gets called when the user clicks on the       *
* "Cancel" button in ask_save_ovl_dialog. It  quits mlab after saving the   *
* overlay name.                                                             *
****************************************************************************/
void save_ovl_cancel_cb(Widget dialog, XtPointer client_data, XtPointer call_data)
{
    char *filename = NULL;

    gEventLogging->log("Do you want to save the waypoints in the overlay file? = Cancel");
    gEventLogging->end("Do you want to save the waypoints in the overlay file?");

    filename = (char *)client_data;

    // Save the overlay file name.
    save_ovlname_to_waypoint_file(filename);

    // Quit mlab.
    quit_mlab();
}

/****************************************************************************
* save_overwrite_ovl                                                        *
*                                                                           *
* This function pops up a dialog which asks the user to see if he/she       *
* really wants to overwrite the existing file.                              *
****************************************************************************/
void ask_overwrite_ovl(char *filename)
{
  Widget ask_overwrite_ovl_dialog = XmCreateQuestionDialog(
      main_window,
      "ask-overwrite-ovl-dialog",
      NULL,
      0);

  XmString yes, no;
  yes = XSTRING("   Yes   ");
  no = XSTRING("   No   ");

  char msg[4096];
  sprintf(msg, "File \"%s\" exists.\nOverwrite it?", filename);
  XmString msg_ = XSTRING(msg);

  XtVaSetValues (ask_overwrite_ovl_dialog,
		 XmNmessageString, msg_,
		 XmNokLabelString, yes,
		 XmNcancelLabelString, no,
		 NULL);

  XtUnmanageChild(XmMessageBoxGetChild(ask_overwrite_ovl_dialog, XmDIALOG_HELP_BUTTON));

  XtAddCallback(ask_overwrite_ovl_dialog, XmNokCallback,
		(XtCallbackProc) overwrite_yes_cb, filename);
  XtAddCallback(ask_overwrite_ovl_dialog, XmNcancelCallback,
		(XtCallbackProc) overwrite_no_cb, filename);

  XmStringFree(yes);
  XmStringFree(no);
  XmStringFree(msg_);

  XtManageChild(ask_overwrite_ovl_dialog);
  
  gEventLogging->start("File %s exists. Overwrite it?", filename);

  return;
}

/****************************************************************************
* save_ovl_as_ok_cb                                                         *
*                                                                           *
* This is a callback function which is called by save_ovl_as_dialog         *
* when the user selects the file name and clicks "Save".                    *
****************************************************************************/
void save_ovl_as_ok_cb(Widget dialog, XtPointer client_data, XtPointer call_data)
{
    char *filename = NULL;
    XmFileSelectionBoxCallbackStruct *cbs = NULL;

    cbs = (XmFileSelectionBoxCallbackStruct *)call_data;

    if (!XmStringGetLtoR(cbs->value, XmFONTLIST_DEFAULT_TAG, &filename))
    {
	warn_userf("Internal Error: Overlay not saved.");
	quit_mlab();
    }

    gEventLogging->log("Save Overlay As = %s", filename);

    XtUnmanageChild(save_ovl_as_dialog);
    gEventLogging->end("Save Overlay As");

    // check to see if the filename needs .ovl extention.
    int len = strlen(filename);
    if (len <= 0)
    {
	warn_userf("Internal Error: Overlay not saved.");
	quit_mlab();
    }

    if (len < 5 || strcmp(&filename[len - 4], ".ovl") != 0)
    {
	// make a copy of the filename with the extension
	char *tmp = (char *) malloc(len + 5);

	strcpy(tmp, filename);
	strcpy(&tmp[len], ".ovl");
	free(filename);
	filename = tmp;
    }

    // Checking if the file exists.
    if (FILE *file = fopen(filename, "r"))
    {
	// It exists. Ask the user if the file should be overwritten.
	fclose(file);
	ask_overwrite_ovl(filename);
    }
    else
    {
	create_new_ovl_for_waypoints(filename);
    }
}

/****************************************************************************
* save_ovl_as_cancel_cb                                                     *
*                                                                           *
* This is a callback function gets called when the user clicks on the       *
* "Cancel" button in save_ovl_as_dialog.                                    *
****************************************************************************/
void save_ovl_as_cancel_cb(Widget dialog, XtPointer client_data, XtPointer call_data)
{
    gEventLogging->log("Save Overlay As = Cancel");

    XtUnmanageChild(save_ovl_as_dialog);
    gEventLogging->end("Save Overlay As");

    // Quit mlab.
    quit_mlab();
}

/****************************************************************************
* create_save_ovl_as_dialog                                                 *
*                                                                           *
* This function creates save_ovl_as_dialog. save_ovl_as_dialog is a dialog  *
* which asks the user to specify the name of the file to save the overlay.  *
****************************************************************************/
void create_save_ovl_as_dialog(Widget parent)
{
  Arg wargs[3];
  int n = 0;

  XmString title = XSTRING("Save Overlay As");
  XmString filterstr = XSTRING("*.ovl");
  XmString save = XSTRING("Save");

  XtSetArg(wargs[n], XmNdialogTitle, title); n++;
  XtSetArg(wargs[n], XmNpattern, filterstr); n++;
  XtSetArg(wargs[n], XmNokLabelString, save); n++;

  save_ovl_as_dialog = XmCreateFileSelectionDialog(parent, "select-file", wargs, n);

  XmStringFree(title);
  XmStringFree(filterstr);
  XmStringFree(save);

  XtUnmanageChild(XmFileSelectionBoxGetChild(save_ovl_as_dialog, XmDIALOG_HELP_BUTTON));

  XtAddCallback(save_ovl_as_dialog,
		XmNokCallback, 
		(XtCallbackProc)save_ovl_as_ok_cb,
		NULL);
  XtAddCallback(save_ovl_as_dialog,
		XmNcancelCallback, 
		(XtCallbackProc)save_ovl_as_cancel_cb,
		NULL);

}

/****************************************************************************
* save_ok_cb                                                                *
*                                                                           *
* This is a callback function invoked by ask_save_ovl_dialog when the user  *
* clicks "Save".                                                            *
****************************************************************************/
void save_ovl_cb(Widget dialog, XtPointer client_data, XtPointer call_data)
{
  char *filename = NULL, *tmp = NULL;
  int len;

  gEventLogging->log("Do you want to save the waypoints in the overlay file? = Save");
  gEventLogging->end("Do you want to save the waypoints in the overlay file?");

  filename = (char *)client_data;

  if (filename == NULL)
  {
      //fprintf(stderr, "Internal Error: Overlay not saved.\n");
      warn_userf("Internal Error: Overlay not saved.");
      quit_mlab(); // internal error
  }

  // check to see if the filename needs .ovl extention.
  len = strlen(filename);

  if (len <= 0)
  {
      warn_userf("Internal Error: Overlay not saved.");
      quit_mlab();
  }

  if (len < 5 || strcmp(&filename[len - 4], ".ovl") != 0)
  {
    // make a copy of the filename with the extension
    tmp = (char *)malloc(len + 5);

    strcpy(tmp, filename);
    strcpy(&tmp[len], ".ovl");
    free(filename);
    filename = tmp;
  }

  // Checking if the file exists.
  if (FILE *file = fopen(filename, "r"))
  {
    // It exists. Ask the user if the file should be overwritten.
    fclose(file);
    ask_overwrite_ovl(filename);
  }
  else
  {
    create_new_ovl_for_waypoints(filename);
  }
}

/****************************************************************************
* ask_save_ovl                                                              *
*                                                                           *
* This is a function which creates and pops up ask_save_ovl_dialog.         *
****************************************************************************/
void ask_save_ovl(void)
{
  XmString save, save_as, cancel, msg;

  ask_save_ovl_dialog = XmCreateQuestionDialog(
      main_window,
      "ask-save-ovl-dialog",
      NULL,
      0);

  save = XSTRING("   Save   ");
  save_as = XSTRING("   Save As   ");
  cancel = XSTRING("   Cancel   ");   
  msg = XSTRING("Do you want to save the waypoints\nin the overlay file?");

  XtVaSetValues (ask_save_ovl_dialog,
		 XmNmessageString, msg,
		 XmNokLabelString, save,
		 XmNcancelLabelString, save_as,
		 XmNhelpLabelString, cancel,
		 NULL);

  XtAddCallback(ask_save_ovl_dialog, XmNokCallback,
		(XtCallbackProc) save_ovl_cb, ovl_filename);
  XtAddCallback(ask_save_ovl_dialog, XmNcancelCallback,
		(XtCallbackProc) save_ovl_as_cb, ovl_filename);
  XtAddCallback(ask_save_ovl_dialog, XmNhelpCallback,
		//tCallbackProc) quit_mlab, NULL);
		(XtCallbackProc) save_ovl_cancel_cb, ovl_filename);

  XmStringFree(save);
  XmStringFree(save_as);
  XmStringFree(cancel);
  XmStringFree(msg);

  /* Pop it up */
  XtManageChild(ask_save_ovl_dialog);

  gEventLogging->start("Do you want to save the waypoints in the overlay file?");
   
  return;
}

/*-----------------------------------------------------------------------*/
/* logs a waypoint to the waypoints file */
void add_waypoint(float x, float y) 
{
  char waypt[1024];
  giNumWayPoints++;
  if (giNumWayPoints > giWayptArySize) {
    /* Too little space so double it */
    giWayptArySize = (giWayptArySize == 0)? 1: giWayptArySize * 2;
    if (!(waypts_ary = (point2d_t*)
	  realloc(waypts_ary, (sizeof(point2d_t)*giWayptArySize)))) {
      fprintf(stderr, "server: out of memory in add_waypoint\n");
      giWayptArySize = 0;
      exit(-1);
    }
  }
  sprintf(waypt, "Waypoint #%d", giNumWayPoints);
  waypts_ary[giNumWayPoints - 1].x = x;
  waypts_ary[giNumWayPoints - 1].y = y;
  gt_add_passage_point(waypt, x, y, 10*meters_per_pixel);
  gEventLogging->log("Waypoing #%d added on <%.2f %.2f>", giNumWayPoints, x, y);
}

/*-----------------------------------------------------------------------*/
void delete_waypoint(float x, float y) 
{
  int i, best;
  double bdist, dist;
  char waypt[1024];

  if (giNumWayPoints == 0) return;

  /* Find the closest waypoint */
  best = 0;
  bdist = (waypts_ary[0].x - x)*(waypts_ary[0].x - x) +
    (waypts_ary[0].y - y)*(waypts_ary[0].y - y);
    
  for (i = 1; i < giNumWayPoints; i++) {
    dist = (waypts_ary[i].x - x)*(waypts_ary[i].x - x) +
      (waypts_ary[i].y - y)*(waypts_ary[i].y - y);
    if (dist < bdist) {
      best = i;
      bdist = dist;
    }
  }
  /* Too far from any waypoint */
  bdist = sqrt(bdist) / meters_per_pixel;
  if (bdist > 25) return;

  /* Delete the waypoint, then renumber all others */
  if (giNumWayPoints == 1) {
    giNumWayPoints = 0;
    sprintf(waypt, "Waypoint #%d", best+1);
    gt_delete_db_entry(waypt);
    gEventLogging->log("Waypoing #%d deleted", best+1);
  }
  else {
    giNumWayPoints--;
    for (i = best; i < giNumWayPoints; i++) {
      waypts_ary[i].x = waypts_ary[i + 1].x;
      waypts_ary[i].y = waypts_ary[i + 1].y;  
      sprintf(waypt, "Waypoint #%d", i+1);
      gt_delete_db_entry(waypt);
      gt_add_passage_point(waypt, waypts_ary[i].x, 
			   waypts_ary[i].y, 10*meters_per_pixel);
      gEventLogging->log("Renumbering Waypoing #%d to #%d", i+2, i+1);
    }
    sprintf(waypt, "Waypoint #%d", i+1);
    gt_delete_db_entry(waypt);
  }
  
  /* Clear the screen and redraw */
  clear_map();
  draw_world();
}

/*-----------------------------------------------------------------------*/
/* exits missionlab */
void return_waypoints(void) 
{
    FILE *fp;
    int i;

    if (!(fp = fopen(waypoints_filename, "w")))
    {
	fprintf(stderr, "Can't write to waypoints file \"%s\".\n", waypoints_filename);
	quit_mlab();
    }

    for (i = 0; i < giNumWayPoints; i++)
    {
	fprintf(fp, "waypoint %f %f\n", waypts_ary[i].x, waypts_ary[i].y);
    }

    fclose(fp);
    free(waypts_ary);
    ask_save_ovl();
}

/*-----------------------------------------------------------------------*/
void save_ovlname_to_waypoint_file(char *ovlname)
{
    FILE *wptfile;
    char simpleFilename[1024];

    if (ovlname == NULL) return;

    if (filename_has_directory(ovlname))
    {
	remove_directory(ovlname, simpleFilename);
	free(ovlname);
	ovlname = strdup(simpleFilename);
    }

    wptfile = fopen(waypoints_filename, "a");

    if (wptfile == NULL)
    {
	fprintf(stderr, "Warning(cfgedit): save_ovlname_to_waypoint_file() failed.\n");
	return;
    }

    fprintf(wptfile, "\noverlay %s", ovlname);
    fclose(wptfile);
    return;
}

/**********************************************************************
 * $Log: waypoint.c,v $
 * Revision 1.1.1.1  2008/07/14 16:44:22  endo
 * MAST Project (based on MissionLab-MINOS-20071018.tar.gz)
 *
 * Revision 1.1.1.1  2006/07/20 17:17:50  endo
 * MINOS Project (based on MissionLab-7.0.20060712.tar.gz)
 *
 * Revision 1.1.1.1  2006/07/12 13:37:59  endo
 * MissionLab 7.0
 *
 * Revision 1.1.1.1  2005/02/06 23:00:11  endo
 * AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
 *
 * Revision 1.2  2002/01/12 23:00:58  endo
 * Mission Expert functionality added.
 *
 * Revision 1.1  2000/07/07 18:26:59  endo
 * Initial revision
 *
 **********************************************************************/
